/* libtinymail-camel_bs - The Tiny Mail base library for CamelBs
 * Copyright (C) 2006-2007 Philip Van Hoof <pvanhoof@gnome.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with self library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#include <config.h>

#include <glib/gi18n-lib.h>

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <string.h>

#include <tny-mime-part.h>
#include <tny-camel-shared.h>
#include <tny-list.h>
#include <tny-camel-bs-msg.h>

#include <tny-fs-stream.h>

static GObjectClass *parent_class = NULL;

#include "tny-camel-bs-mime-part-priv.h"
#include "tny-camel-bs-msg-priv.h"


static void 
tny_camel_bs_mime_part_set_header_pair (TnyMimePart *self, const gchar *name, const gchar *value)
{
	TNY_CAMEL_BS_MIME_PART_GET_CLASS (self)->set_header_pair_func (self, name, value);
	return;
}

static void 
tny_camel_bs_mime_part_set_header_pair_default (TnyMimePart *self, const gchar *name, const gchar *value)
{
	printf ("Not supported\n");
}

static void 
tny_camel_bs_mime_part_get_header_pairs (TnyMimePart *self, TnyList *list)
{
	TNY_CAMEL_BS_MIME_PART_GET_CLASS (self)->get_header_pairs_func (self, list);
	return;
}

static void 
tny_camel_bs_mime_part_get_header_pairs_default (TnyMimePart *self, TnyList *list)
{
	printf ("Not supported");
	return;
}

static void
tny_camel_bs_mime_part_get_parts (TnyMimePart *self, TnyList *list)
{
	TNY_CAMEL_BS_MIME_PART_GET_CLASS (self)->get_parts_func (self, list);
	return;
}

static void
tny_camel_bs_mime_part_get_parts_default (TnyMimePart *self, TnyList *list)
{
	TnyCamelBsMimePartPriv *priv = TNY_CAMEL_BS_MIME_PART_GET_PRIVATE (self);
	bodystruct_t *part = priv->bodystructure;

	g_mutex_lock (priv->part_lock);

	if (part->content.type && !strcasecmp (part->content.type, "multipart")) {
		part = part->subparts;
		while (part != NULL) {
			TnyMimePart *mpart = _tny_camel_bs_mime_part_new (part, priv->cache_path);
			tny_list_prepend (list, (GObject *) mpart);
			g_object_unref (mpart);
			part = part->next;
		}
	}

	g_mutex_unlock (priv->part_lock);

	return;
}



static gint
tny_camel_bs_mime_part_add_part (TnyMimePart *self, TnyMimePart *part)
{
	return TNY_CAMEL_BS_MIME_PART_GET_CLASS (self)->add_part_func (self, part);
}

static gint
tny_camel_bs_mime_part_add_part_default (TnyMimePart *self, TnyMimePart *part)
{
	printf ("Not supported");
	return -1;
}


static void 
tny_camel_bs_mime_part_del_part (TnyMimePart *self,  TnyMimePart *part)
{
	TNY_CAMEL_BS_MIME_PART_GET_CLASS (self)->del_part_func (self, part);
	return;
}

static void 
tny_camel_bs_mime_part_del_part_default (TnyMimePart *self, TnyMimePart *part)
{
	printf ("Not supported");
	return;
}



static gboolean 
tny_camel_bs_mime_part_is_attachment (TnyMimePart *self)
{
	return TNY_CAMEL_BS_MIME_PART_GET_CLASS (self)->is_attachment_func (self);
}


static gboolean 
tny_camel_bs_mime_part_is_attachment_default (TnyMimePart *self)
{
	TnyCamelBsMimePartPriv *priv = TNY_CAMEL_BS_MIME_PART_GET_PRIVATE (self);
	const gchar *contdisp = priv->bodystructure->disposition.type;

	/* Content-Disposition is excellent for this, of course (but we might
	 * not actually have this header, as not all E-mail clients add it) */

	if (contdisp) {
		if (camel_strstrcase (contdisp, "inline"))
			return FALSE;
		if (camel_strstrcase (contdisp, "attachment"))
			return TRUE;
	}

	return !((camel_strcase_equal (priv->bodystructure->content.type, "application") && 
		 camel_strcase_equal (priv->bodystructure->content.subtype, "x-pkcs7-mime")) ||
		(camel_strcase_equal (priv->bodystructure->content.type, "application") && 
 		camel_strcase_equal (priv->bodystructure->content.subtype, "x-pkcs7-mime")) ||
		(camel_strcase_equal (priv->bodystructure->content.type, "application") && 
 		camel_strcase_equal (priv->bodystructure->content.subtype, "x-inlinepgp-signed")) ||
		(camel_strcase_equal (priv->bodystructure->content.type, "application") && 
 		camel_strcase_equal (priv->bodystructure->content.subtype, "x-inlinepgp-encrypted")) ||
		(mimeparam_get_value_for (priv->bodystructure->disposition.params, "FILENAME") == NULL));

}

static void
tny_camel_bs_mime_part_write_to_stream (TnyMimePart *self, TnyStream *stream)
{
	TNY_CAMEL_BS_MIME_PART_GET_CLASS (self)->write_to_stream_func (self, stream);
	return;
}

static void
tny_camel_bs_mime_part_write_to_stream_default (TnyMimePart *self, TnyStream *stream)
{
	TnyCamelBsMimePartPriv *priv = TNY_CAMEL_BS_MIME_PART_GET_PRIVATE (self);

	return;
}



static void
tny_camel_bs_mime_part_decode_to_stream (TnyMimePart *self, TnyStream *stream)
{
	TNY_CAMEL_BS_MIME_PART_GET_CLASS (self)->decode_to_stream_func (self, stream);
	return;
}

static void
tny_camel_bs_mime_part_decode_to_stream_default (TnyMimePart *self, TnyStream *stream)
{

	TnyCamelBsMimePartPriv *priv = TNY_CAMEL_BS_MIME_PART_GET_PRIVATE (self);
	gchar *filename = g_strdup_printf ("%s/%s", priv->cache_path, priv->bodystructure->imap_id);
	int fd = open (filename, 0);

	if (fd == -1) {
		printf ("Get %s from server\n", priv->bodystructure->imap_id);
		fd = open (filename, 0);
	}

	if (fd != -1) {
		TnyStream *fs_stream = tny_fs_stream_new (fd);
		tny_stream_write_to_stream (fs_stream, stream);
	}

	return;
}

static gint
tny_camel_bs_mime_part_construct_from_stream (TnyMimePart *self, TnyStream *stream, const gchar *type)
{
	return TNY_CAMEL_BS_MIME_PART_GET_CLASS (self)->construct_from_stream_func (self, stream, type);
}

static gint
tny_camel_bs_mime_part_construct_from_stream_default (TnyMimePart *self, TnyStream *stream, const gchar *type)
{
	TnyCamelBsMimePartPriv *priv = TNY_CAMEL_BS_MIME_PART_GET_PRIVATE (self);
	return -1;
}

static TnyStream* 
tny_camel_bs_mime_part_get_stream (TnyMimePart *self)
{
	return TNY_CAMEL_BS_MIME_PART_GET_CLASS (self)->get_stream_func (self);
}

static TnyStream* 
tny_camel_bs_mime_part_get_stream_default (TnyMimePart *self)
{
	TnyCamelBsMimePartPriv *priv = TNY_CAMEL_BS_MIME_PART_GET_PRIVATE (self);
	TnyStream *retval = NULL;

	return retval;
}

static const gchar* 
tny_camel_bs_mime_part_get_content_type (TnyMimePart *self)
{
	return TNY_CAMEL_BS_MIME_PART_GET_CLASS (self)->get_content_type_func (self);
}

static const gchar* 
tny_camel_bs_mime_part_get_content_type_default (TnyMimePart *self)
{
	TnyCamelBsMimePartPriv *priv = TNY_CAMEL_BS_MIME_PART_GET_PRIVATE (self);

	if (G_LIKELY (!priv->cached_content_type)) {
		g_mutex_lock (priv->part_lock);
		priv->cached_content_type = g_strdup_printf ("%s/%s", 
			priv->bodystructure->content.type, 
			priv->bodystructure->content.subtype);
		g_mutex_unlock (priv->part_lock);
	}

	return priv->cached_content_type;
}

static gboolean
tny_camel_bs_mime_part_is_purged (TnyMimePart *self)
{
	return TNY_CAMEL_BS_MIME_PART_GET_CLASS (self)->is_purged_func (self);
}

static gboolean
tny_camel_bs_mime_part_is_purged_default (TnyMimePart *self)
{
	TnyCamelBsMimePartPriv *priv = TNY_CAMEL_BS_MIME_PART_GET_PRIVATE (self);

	printf ("Not supported\n");

	return FALSE;
}

static gboolean 
tny_camel_bs_mime_part_content_type_is (TnyMimePart *self, const gchar *type)
{
	return TNY_CAMEL_BS_MIME_PART_GET_CLASS (self)->content_type_is_func (self, type);
}

static gboolean 
tny_camel_bs_mime_part_content_type_is_default (TnyMimePart *self, const gchar *type)
{
	TnyCamelBsMimePartPriv *priv = TNY_CAMEL_BS_MIME_PART_GET_PRIVATE (self);
	gchar *dup, *str1, *str2, *ptr;
	gboolean retval = FALSE;

	/* Whoooo, pointer hocus .. */

	dup = g_strdup (type);
	ptr = strchr (dup, '/');
	ptr++; str2 = g_strdup (ptr);
	ptr--; *ptr = '\0'; str1 = dup;

	/* pocus ! */

	if (!strcmp (str2, "*")) {
		if (camel_strcase_equal (priv->bodystructure->content.type, str1))
				retval = TRUE;
	} else {
		if (camel_strcase_equal (priv->bodystructure->content.type, str1) &&
			camel_strcase_equal (priv->bodystructure->content.subtype, str2))
				retval = TRUE;
	}

	g_free (dup);
	g_free (str2);

	return retval;
}


static const gchar*
tny_camel_bs_mime_part_get_filename (TnyMimePart *self)
{
	return TNY_CAMEL_BS_MIME_PART_GET_CLASS (self)->get_filename_func (self);
}

static const gchar*
tny_camel_bs_mime_part_get_filename_default (TnyMimePart *self)
{
	TnyCamelBsMimePartPriv *priv = TNY_CAMEL_BS_MIME_PART_GET_PRIVATE (self);
	const gchar *retval;

	g_mutex_lock (priv->part_lock);
	retval = mimeparam_get_value_for (priv->bodystructure->disposition.params, "FILENAME");
	g_mutex_unlock (priv->part_lock);

	return retval;
}

static const gchar*
tny_camel_bs_mime_part_get_content_id (TnyMimePart *self)
{
	return TNY_CAMEL_BS_MIME_PART_GET_CLASS (self)->get_content_id_func (self);
}

static const gchar*
tny_camel_bs_mime_part_get_content_id_default (TnyMimePart *self)
{
	TnyCamelBsMimePartPriv *priv = TNY_CAMEL_BS_MIME_PART_GET_PRIVATE (self);
	const gchar *retval;

	g_mutex_lock (priv->part_lock);
	retval = priv->bodystructure->content.cid;
	g_mutex_unlock (priv->part_lock);

	return retval;
}

static const gchar*
tny_camel_bs_mime_part_get_description (TnyMimePart *self)
{
	return TNY_CAMEL_BS_MIME_PART_GET_CLASS (self)->get_description_func (self);
}

static const gchar*
tny_camel_bs_mime_part_get_description_default (TnyMimePart *self)
{
	TnyCamelBsMimePartPriv *priv = TNY_CAMEL_BS_MIME_PART_GET_PRIVATE (self);
	const gchar *retval;

	g_mutex_lock (priv->part_lock);
	retval = priv->bodystructure->description;
	g_mutex_unlock (priv->part_lock);

	return retval;
}

static const gchar*
tny_camel_bs_mime_part_get_content_location (TnyMimePart *self)
{
	return TNY_CAMEL_BS_MIME_PART_GET_CLASS (self)->get_content_location_func (self);
}

static const gchar*
tny_camel_bs_mime_part_get_content_location_default (TnyMimePart *self)
{
	TnyCamelBsMimePartPriv *priv = TNY_CAMEL_BS_MIME_PART_GET_PRIVATE (self);
	const gchar *retval;

	g_mutex_lock (priv->part_lock);
	retval = priv->bodystructure->content.loc;
	g_mutex_unlock (priv->part_lock);

	return retval;
}


static void 
tny_camel_bs_mime_part_set_content_location (TnyMimePart *self, const gchar *content_location)
{
	TNY_CAMEL_BS_MIME_PART_GET_CLASS (self)->set_content_location_func (self, content_location);
	return;
}

static void 
tny_camel_bs_mime_part_set_content_location_default (TnyMimePart *self, const gchar *content_location)
{
	printf ("Not supported\n");
}

static void 
tny_camel_bs_mime_part_set_description (TnyMimePart *self, const gchar *description)
{
	TNY_CAMEL_BS_MIME_PART_GET_CLASS (self)->set_description_func (self, description);
	return;
}

static void 
tny_camel_bs_mime_part_set_description_default (TnyMimePart *self, const gchar *description)
{
	printf ("Not supported\n");
}

static void 
tny_camel_bs_mime_part_set_content_id (TnyMimePart *self, const gchar *content_id)
{
	TNY_CAMEL_BS_MIME_PART_GET_CLASS (self)->set_content_id_func (self, content_id);
	return;
}

static void 
tny_camel_bs_mime_part_set_content_id_default (TnyMimePart *self, const gchar *content_id)
{
	printf ("Not supported\n");
}

static void 
tny_camel_bs_mime_part_set_filename (TnyMimePart *self, const gchar *filename)
{
	TNY_CAMEL_BS_MIME_PART_GET_CLASS (self)->set_filename_func (self, filename);
	return;
}

static void 
tny_camel_bs_mime_part_set_filename_default (TnyMimePart *self, const gchar *filename)
{
	printf ("Not supported\n");
}


static void 
tny_camel_bs_mime_part_set_content_type (TnyMimePart *self, const gchar *content_type)
{
	TNY_CAMEL_BS_MIME_PART_GET_CLASS (self)->set_content_type_func (self, content_type);
	return;
}

static void
tny_camel_bs_mime_part_set_purged_default (TnyMimePart *self)
{
	printf ("Not supported\n");
}

static void
tny_camel_bs_mime_part_set_purged (TnyMimePart *self)
{
	TNY_CAMEL_BS_MIME_PART_GET_CLASS (self)->set_purged_func (self);
	return;
}

static void 
tny_camel_bs_mime_part_set_content_type_default (TnyMimePart *self, const gchar *content_type)
{
	printf ("Not supported\n");

	return;
}

static void
tny_camel_bs_mime_part_finalize (GObject *object)
{
	TnyCamelBsMimePart *self = (TnyCamelBsMimePart*) object;
	TnyCamelBsMimePartPriv *priv = TNY_CAMEL_BS_MIME_PART_GET_PRIVATE (self);

	g_mutex_lock (priv->part_lock);
	if (priv->cached_content_type)
		g_free (priv->cached_content_type);
	priv->cached_content_type = NULL;
	g_mutex_unlock (priv->part_lock);

	g_mutex_free (priv->part_lock);
	priv->part_lock = NULL;

	g_free (priv->cache_path);
	bodystruct_free (priv->bodystructure);

	(*parent_class->finalize) (object);

	return;
}

TnyMimePart*
_tny_camel_bs_mime_part_new (bodystruct_t *bodystructure, const gchar *cache_path)
{
	TnyCamelBsMimePart *self = g_object_new (TNY_TYPE_CAMEL_BS_MIME_PART, NULL);
	TnyCamelBsMimePartPriv *priv = TNY_CAMEL_BS_MIME_PART_GET_PRIVATE (self);

	priv->bodystructure = bodystructure;
	priv->cache_path = g_strdup (cache_path);

	return TNY_MIME_PART (self);
}


static void
tny_mime_part_init (gpointer g, gpointer iface_data)
{
	TnyMimePartIface *klass = (TnyMimePartIface *)g;

	klass->content_type_is_func = tny_camel_bs_mime_part_content_type_is;
	klass->get_content_type_func = tny_camel_bs_mime_part_get_content_type;
	klass->get_stream_func = tny_camel_bs_mime_part_get_stream;
	klass->write_to_stream_func = tny_camel_bs_mime_part_write_to_stream;
	klass->construct_from_stream_func = tny_camel_bs_mime_part_construct_from_stream;
	klass->get_filename_func = tny_camel_bs_mime_part_get_filename;
	klass->get_content_id_func = tny_camel_bs_mime_part_get_content_id;
	klass->get_description_func = tny_camel_bs_mime_part_get_description;
	klass->get_content_location_func = tny_camel_bs_mime_part_get_content_location;
	klass->is_purged_func = tny_camel_bs_mime_part_is_purged;
	klass->set_content_location_func = tny_camel_bs_mime_part_set_content_location;
	klass->set_description_func = tny_camel_bs_mime_part_set_description;
	klass->set_purged_func = tny_camel_bs_mime_part_set_purged;
	klass->set_content_id_func = tny_camel_bs_mime_part_set_content_id;
	klass->set_filename_func = tny_camel_bs_mime_part_set_filename;
	klass->set_content_type_func = tny_camel_bs_mime_part_set_content_type;
	klass->is_attachment_func = tny_camel_bs_mime_part_is_attachment;
	klass->decode_to_stream_func = tny_camel_bs_mime_part_decode_to_stream;
	klass->get_parts_func = tny_camel_bs_mime_part_get_parts;
	klass->add_part_func = tny_camel_bs_mime_part_add_part;
	klass->del_part_func = tny_camel_bs_mime_part_del_part;
	klass->get_header_pairs_func = tny_camel_bs_mime_part_get_header_pairs;
	klass->set_header_pair_func = tny_camel_bs_mime_part_set_header_pair;

	return;
}


static void 
tny_camel_bs_mime_part_class_init (TnyCamelBsMimePartClass *class)
{
	GObjectClass *object_class;

	parent_class = g_type_class_peek_parent (class);
	object_class = (GObjectClass*) class;

	class->content_type_is_func = tny_camel_bs_mime_part_content_type_is_default;
	class->get_content_type_func = tny_camel_bs_mime_part_get_content_type_default;
	class->get_stream_func = tny_camel_bs_mime_part_get_stream_default;
	class->write_to_stream_func = tny_camel_bs_mime_part_write_to_stream_default;
	class->construct_from_stream_func = tny_camel_bs_mime_part_construct_from_stream_default;
	class->get_filename_func = tny_camel_bs_mime_part_get_filename_default;
	class->get_content_id_func = tny_camel_bs_mime_part_get_content_id_default;
	class->get_description_func = tny_camel_bs_mime_part_get_description_default;
	class->get_content_location_func = tny_camel_bs_mime_part_get_content_location_default;
	class->is_purged_func = tny_camel_bs_mime_part_is_purged_default;
	class->set_purged_func = tny_camel_bs_mime_part_set_purged_default;
	class->set_content_location_func = tny_camel_bs_mime_part_set_content_location_default;
	class->set_description_func = tny_camel_bs_mime_part_set_description_default;
	class->set_content_id_func = tny_camel_bs_mime_part_set_content_id_default;
	class->set_filename_func = tny_camel_bs_mime_part_set_filename_default;
	class->set_content_type_func = tny_camel_bs_mime_part_set_content_type_default;
	class->is_attachment_func = tny_camel_bs_mime_part_is_attachment_default;
	class->decode_to_stream_func = tny_camel_bs_mime_part_decode_to_stream_default;
	class->get_parts_func = tny_camel_bs_mime_part_get_parts_default;
	class->add_part_func = tny_camel_bs_mime_part_add_part_default;
	class->del_part_func = tny_camel_bs_mime_part_del_part_default;
	class->get_header_pairs_func = tny_camel_bs_mime_part_get_header_pairs_default;
	class->set_header_pair_func = tny_camel_bs_mime_part_set_header_pair_default;

	object_class->finalize = tny_camel_bs_mime_part_finalize;

	g_type_class_add_private (object_class, sizeof (TnyCamelBsMimePartPriv));

	return;
}

static void
tny_camel_bs_mime_part_instance_init (GTypeInstance *instance, gpointer g_class)
{
	TnyCamelBsMimePart *self = (TnyCamelBsMimePart*)instance;
	TnyCamelBsMimePartPriv *priv = TNY_CAMEL_BS_MIME_PART_GET_PRIVATE (self);

	priv->bodystructure = NULL;
	priv->part_lock = g_mutex_new ();

	return;
}

/**
 * tny_camel_bs_mime_part_get_type:
 *
 * GType system helper function
 *
 * Return value: a GType
 **/
GType 
tny_camel_bs_mime_part_get_type (void)
{
	static GType type = 0;

	if (G_UNLIKELY (!_camel_type_init_done))
	{
		if (!g_thread_supported ()) 
			g_thread_init (NULL);

		camel_type_init ();
		_camel_type_init_done = TRUE;
	}

	if (G_UNLIKELY(type == 0))
	{
		static const GTypeInfo info = 
		{
		  sizeof (TnyCamelBsMimePartClass),
		  NULL,   /* base_init */
		  NULL,   /* base_finalize */
		  (GClassInitFunc) tny_camel_bs_mime_part_class_init,   /* class_init */
		  NULL,   /* class_finalize */
		  NULL,   /* class_data */
		  sizeof (TnyCamelBsMimePart),
		  0,      /* n_preallocs */
		  tny_camel_bs_mime_part_instance_init,    /* instance_init */
		  NULL
		};

		static const GInterfaceInfo tny_mime_part_info = 
		{
		  (GInterfaceInitFunc) tny_mime_part_init, /* interface_init */
		  NULL,         /* interface_finalize */
		  NULL          /* interface_data */
		};

		type = g_type_register_static (G_TYPE_OBJECT,
			"TnyCamelBsMimePart",
			&info, 0);

		g_type_add_interface_static (type, TNY_TYPE_MIME_PART, 
			&tny_mime_part_info);
	}

	return type;
}
